/* ----------------------------------------------------------------------------
  BTRIEVE LOG FILE SIMPLE VIEWER                                        G.A.M.
  ----------------------------------------------------------------------------*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <io.h>
#include <dos.h>
#include <sys\stat.h>
#include <time.h>

#define BlogSign "BLOG"
#define BlogSignSize          4
#define BlogCreateTimeOffset 40
#define RecDescSize          32

char *opNames[] = {
       "OPEN"    ,
       "CLOSE"   ,
       "INSERT"  ,	
       "UPDATE"  ,	
       "DELETE"  ,	// without data buffer :(
       "GETEQ"   ,		
       "NEXT"    ,		
       "PREV"    ,		
       "GETGT",		
       "GETGE",		
       "GETLT",		
       "GETLE",		
       "FIRST",		
       "LAST" ,		
       "CREATE",	
       "STAT"  ,		
       "EXTEND",	
       "SETDIR",	
       "GETDIR",	
       "BEGIN TRANS." ,		
       "END TRANS."   ,		
       "ABORT TRANS.",
       "GETPOS"	,
       "GETDIRECT",	
       "STEPNEXT" ,	
       "STOP"	  ,	
       "VERSION"  ,	
       "UNLOCK"	  ,
       "RESET"	  ,	
       "SETOWNER" ,	
       "CLROWNER" ,	
       "CREIND"	  ,
       "DROPIND"  ,	
       "STEPFIRST",	
       "STEPLAST" ,	
       "STEPPREV",
       NULL
    };		
int opNamesSize = 0;

char *month[] = {"---", "Jan", "Feb", "Mar", "Apr", "May", "Jun",
                 "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
/*
       "BLOG"         blog file mark
      +40  2          file create time (UNIX style)
      ---------------------------------         RECORD DESCRIPTOR
          0   2          btrieve operation code
         +2   1          key 
         +3   4          position
         +7  10          net address
         +17  3          ??
         +20  2          data size
         +22  2          ??
         +24  4          operation date&time (UNIX style)
         +28  1          user name len
         +29  2          key data buffer len
         +31  7          ???????
      ----------------------------------        DATA
         +38   <28>       user name
         +<28> <20>       data 
         +<20> <29>       key
      -----------------------------------     .... next record
*/

#pragma pack(0)
struct  {

     unsigned int   cOp;
     unsigned char  key;
     unsigned long  position;
     unsigned char  netadress[10];
     unsigned char  reserv_1[3];
     unsigned int   DataSize;
     unsigned char  reserv_2[2];
     unsigned long  time;
     unsigned char  UserNameSize;
     unsigned int   KeySize;
     unsigned char  reserv_3[7];

} recdesc;



#define HEXGROUPDELTA     1
#define HEXCHARBYGROUP    4
#define HEXTOTALGROUP     4
#define HEXCHAR           2
#define HEXCHARDELTA      1

#define HEXGROUPWIDTH     (HEXCHARBYGROUP * (HEXCHARDELTA+HEXCHAR))
#define HEXWIDTH          (HEXGROUPWIDTH * HEXTOTALGROUP)
#define HEXFULLWIDTH      (HEXWIDTH + (HEXGROUPDELTA*HEXTOTALGROUP))

#define LINESEPARATOR     "\n "

void dumpstr(unsigned char *beg,unsigned  char *end)
{
  while (beg != end)
  {
    putchar((*beg) < ' ' ? '.' : *beg);
    beg++;
  }
}

void dumprint(unsigned char *data, int size)
{
 unsigned char *text, buffer[(HEXTOTALGROUP*HEXCHARBYGROUP)+2];
 unsigned int ip = 0, i;

  printf(LINESEPARATOR);
  text = buffer;
  for (i =0; i < size; i++)
  {
     ip += printf("%*.*X",HEXCHAR,HEXCHAR, *data);
     ip += printf("%*s",HEXCHARDELTA,"");
     if (!(ip % HEXGROUPWIDTH))
	  printf("%*s",HEXGROUPDELTA,"");
     *text++ = *data;
     ip %= HEXWIDTH;
     if (!ip)  // eol state
     {
	dumpstr(buffer,text);
	ip = 0;
	text = buffer;
	printf(LINESEPARATOR);
     }
     data++;
  }

  if (text != buffer)
  {
      i = HEXFULLWIDTH - ip - (ip/HEXGROUPWIDTH)*HEXGROUPDELTA;
      printf("%*s",i,"");
      dumpstr(buffer,text);
  }

}

char *getBtrOperationName( int code )
{
 static char tmp[8];
 return ( code < opNamesSize ? opNames[code] : itoa(code,tmp,10) );

}

int main( )
{
 int handle, i;
 char sign[ BlogSignSize + 1 ];
/* char recdesc[ RecDescSize + 1 ]; */
 unsigned long BlogCreateTime, offs;
 struct date dt;
 struct time tm;
 char  *userName, *Data, *Key;

 while (opNames[++opNamesSize]) ;

  if (_argc < 2 ) 
  {
    printf("\nBLOG FILE NAME NEED!");
    return 1;
  }

  if ((handle = open(_argv[1], O_RDONLY | O_BINARY)) == -1)
  {
    printf("\nCAN'T OPEN FILE %s",_argv[1]);
    return 1;
  }

  if (     ( read(handle, sign, BlogSignSize) < BlogSignSize )
	|| (strncmp(sign,BlogSign,BlogSignSize)) )
  {
    printf("\n Not Btrieve LOG file!");
    return 2;
  }

  if ( lseek( handle, BlogCreateTimeOffset, SEEK_SET) < BlogCreateTimeOffset )
  {
    printf("\n Not Btrieve LOG file!");
    return 2;
  }

  if ( read(handle, &BlogCreateTime, sizeof(BlogCreateTime)) < 0 )
  {
      printf("\n BLOG file corrupted!");
    return 3;
  }
  printf("\n\n---< %s >-----------------------------------------------:",_argv[1]);
  unixtodos(BlogCreateTime, &dt, &tm);
  printf("\nCreate time %d/%d/%d  %.2u:%.2u:%.2u",
     dt.da_day, dt.da_mon, dt.da_year, 
     tm.ti_hour,tm.ti_min,tm.ti_sec);
  

  while (!eof(handle))
  {
    if ( read(handle, &recdesc, sizeof(recdesc)) < sizeof(recdesc) )
    {
      printf("\n BLOG file corrupted!");
      return 3;
    }          
    printf("\n Operation: %-15.15s Net: ",
	       getBtrOperationName(recdesc.cOp));
    for (i=0; i < 10; i++)
      printf("%x", recdesc.netadress[i] );
    unixtodos(recdesc.time, &dt, &tm);
    printf(" Time: %.2d/%.2d/%d %.2u:%.2u:%.2u",
       dt.da_day, dt.da_mon, dt.da_year, 
       tm.ti_hour,tm.ti_min,tm.ti_sec);

    userName = malloc(recdesc.UserNameSize+1);
    if (userName)
    {
      if ( read( handle, userName, recdesc.UserNameSize) < 0 )
      {
	printf("\n BLOG file corrupted!");
	return 3;
      }
      userName[ recdesc.UserNameSize ] =  '\0';
      printf("\n User : %s", userName);
      free(userName);
      userName = NULL;
    }

    if ( (recdesc.DataSize > 0) && ((Data = malloc(recdesc.DataSize))!=NULL) )
    {
      if ( (i=read( handle, Data, recdesc.DataSize)) <= 0 )
      {
	printf("\n BLOG file corrupted!");
	return 3;
      }
      printf("\n Data :");
      dumprint(Data, i);
      free(Data);
      Data = NULL;
    }

    if ( (recdesc.KeySize> 0) && ((Key = malloc(recdesc.KeySize))!=NULL) )
    {
      if ( (i=read( handle, Key, recdesc.KeySize)) <= 0 )
      {
	printf("\n BLOG corrupted!");
	return 3;
      }
      printf("\n Key %u :", recdesc.key);
      dumprint(Key, i);

      free(Key);
      Key = NULL;
    }

  }
  return 0;
}

